
#include <iostream>

#include <stdlib.h>

#include <SDL.h>

#ifdef TEST
    #include <SDL_thread.h>
#endif /* TEST */

#include "streamPipe.h"


StreamPipe::StreamPipe( unsigned int _maxBufLen )
    : maxBufLen( _maxBufLen+1 ),
	curWriteIndex( 0 ), curReadIndex( 0 ),
	writeIsClosed( false ), readIsClosed( false )
{
    this->buffer = new unsigned char[ this->maxBufLen ];

    this->mutex = ::SDL_CreateMutex();
    this->writeBytesAvailable = ::SDL_CreateCond();
    this->readBytesAvailable = ::SDL_CreateCond();
}

StreamPipe::~StreamPipe( void )
{
    ::SDL_DestroyCond( this->readBytesAvailable );
    ::SDL_DestroyCond( this->writeBytesAvailable );
    ::SDL_DestroyMutex( this->mutex );

    delete[] this->buffer;
}

int
StreamPipe::writeBytesCount( void ) const
{
    if ( this->readIsClosed || this->writeIsClosed ) {
	return -1;
    } else if ( this->curWriteIndex < this->curReadIndex ) {
	return this->curReadIndex - this->curWriteIndex - 1;
    } else {
	return this->maxBufLen - this->curWriteIndex + this->curReadIndex - 1;
    }
}

int
StreamPipe::readBytesCount( void ) const
{
    int count;

    if ( this->readIsClosed ) {
	count = -1;
    } else if ( this->curWriteIndex >= this->curReadIndex ) {
	count = this->curWriteIndex - this->curReadIndex;

	if ( count == 0 && this->writeIsClosed ) {
	    count = -1;
	}
    } else {
	count = this->maxBufLen - this->curReadIndex + this->curWriteIndex;
    }

    return count;
}

int
StreamPipe::write( const void* buf, unsigned int len )
{
    const unsigned char* bufPtr = (const unsigned char*)buf;
    int ret = 0;

    while ( len > 0 ) {
	int rawRet = this->rawWrite( bufPtr, len );

	if ( rawRet <= 0 ) {
	    break;
	} else {
	    bufPtr = &bufPtr[ rawRet ];
	    ret += rawRet;
	    len -= rawRet;
	}
    }

    return ret;
}

int
StreamPipe::read( void* buf, unsigned int len )
{
    unsigned char* bufPtr = (unsigned char*)buf;
    int ret = 0;

    while ( len > 0 ) {
	int rawRet = this->rawRead( bufPtr, len );

	if ( rawRet <= 0 ) {
	    break;
	} else {
	    bufPtr = &bufPtr[ rawRet ];
	    ret += rawRet;
	    len -= rawRet;
	}
    }

    return ret;
}

void
StreamPipe::writeClose( void )
{
    this->writeIsClosed = true;
}

void
StreamPipe::readClose( void )
{
    this->readIsClosed = true;
}

int
StreamPipe::rawWrite( const void* buf, unsigned int len )
{
    const unsigned char* bufPtr = (const unsigned char*)buf;

    ::SDL_LockMutex( this->mutex );

	int wbCount;

	for (;;) {
	    wbCount = this->writeBytesCount();

	    if ( wbCount == 0 ) {
		::SDL_CondWait( this->writeBytesAvailable, this->mutex );
	    } else {
		break;
	    }
	}

	if ( wbCount < 0 ) {
	    return wbCount;
	} else if ( wbCount < len ) {
	    len = (unsigned int)wbCount;
	}

	if ( this->curWriteIndex + len <= this->maxBufLen ) {
	    ::memcpy( &this->buffer[ this->curWriteIndex ], &bufPtr[ 0 ], len );
	    this->curWriteIndex += len;
	    if ( this->curWriteIndex == this->maxBufLen ) {
		this->curWriteIndex = 0;
	    }
	} else {
	    unsigned int endLen = this->maxBufLen - this->curWriteIndex;
	    ::memcpy(
		&this->buffer[ this->curWriteIndex ], &bufPtr[ 0 ], endLen
	    );
	    ::memcpy( &this->buffer[ 0 ], &bufPtr[ endLen ], len - endLen );
	    this->curWriteIndex = len - endLen;
	}

	::SDL_CondSignal( this->readBytesAvailable );

    ::SDL_UnlockMutex( this->mutex );

    return (int)len;
}

int
StreamPipe::rawRead( void* buf, unsigned int len )
{
    unsigned char* bufPtr = (unsigned char*)buf;

    ::SDL_LockMutex( this->mutex );

	int rbCount;

	for (;;) {
	    rbCount = this->readBytesCount();

	    if ( rbCount == 0 ) {
		::SDL_CondWait( this->readBytesAvailable, this->mutex );
	    } else {
		break;
	    }
	}

	if ( rbCount < 0 ) {
	    return rbCount;
	} else if ( rbCount < len ) {
	    len = (unsigned int)rbCount;
	}

	if ( this->curReadIndex + len <= this->maxBufLen ) {
	    ::memcpy( &bufPtr[ 0 ], &this->buffer[ this->curReadIndex ], len );
	    this->curReadIndex += len;
	    if ( this->curReadIndex == this->maxBufLen ) {
		this->curReadIndex = 0;
	    }
	} else {
	    unsigned int endLen = this->maxBufLen - this->curReadIndex;
	    ::memcpy(
		&bufPtr[ 0 ], &this->buffer[ this->curReadIndex ], endLen
	    );
	    ::memcpy( &bufPtr[ endLen ], &this->buffer[ 0 ], len - endLen );
	    this->curReadIndex = len - endLen;
	}

	::SDL_CondSignal( this->writeBytesAvailable );

    ::SDL_UnlockMutex( this->mutex );

    return (int)len;
}

#ifdef TEST
typedef struct {
    StreamPipe* pipe;
    unsigned int writeDelay;
    unsigned int readDelay;
    unsigned int writeLen;
    unsigned int readLen;
} MyThreadArgs;

static int
writerFunc( void* data )
{
    MyThreadArgs* args = (MyThreadArgs*)data;

    unsigned int num = 0;

    for ( unsigned int ii=0; ii < 10; ++ii ) {
	if ( args->writeDelay not_eq 0 ) {
	    ::SDL_Delay( args->writeDelay );
	}

	unsigned char buf[ 64 ];

	for ( unsigned int jj=0; jj < args->writeLen; ++jj ) {
	    buf[ jj ] = num++;
	}

	int len = args->pipe->write( buf, args->writeLen );
    }

    args->pipe->writeClose();

    return 0;
}

static int
readerFunc( void* data )
{
    bool ret = true;

    MyThreadArgs* args = (MyThreadArgs*)data;

    unsigned int nextExpected = 0;

    for (;;) {
	if ( args->readDelay not_eq 0 ) {
	    ::SDL_Delay( args->readDelay );
	}

	unsigned char buf[ 64 ];
	int len  = args->pipe->read( buf, args->readLen );

	if ( len < 0 ) {
	    ret = false;
	} else if ( len == 0 ) {
	    if ( nextExpected not_eq args->writeLen * 10 ) {
		std::cerr << "Didn't make it: " << nextExpected << std::endl;
		ret = false;
	    }
	    args->pipe->readClose();
	    return ( ret ) ? 0 : 1;
	} else {
	    for ( unsigned int ii=0; ii < len; ++ii ) {
		if ( buf[ ii ] not_eq nextExpected++ ) {
		    ret = false;
		}
	    }
	}
    }
}

static int
runTest( unsigned int writeDelay, unsigned int readDelay,
    unsigned int writeLen, unsigned int readLen )
{
    MyThreadArgs args;

    StreamPipe pipe( 20 );

    args.pipe = &pipe;
    args.writeDelay = writeDelay;
    args.readDelay = readDelay;
    args.writeLen = writeLen;
    args.readLen = readLen;

    SDL_Thread* reader = ::SDL_CreateThread( readerFunc, &args );
    SDL_Thread* writer = ::SDL_CreateThread( writerFunc, &args );

    int writeStatus;
    int readStatus;

    ::SDL_WaitThread( writer, &writeStatus );
    ::SDL_WaitThread( reader, &readStatus );

    return writeStatus | readStatus;
}

int
StreamPipe::test( void )
{
    int ret;

    ret = runTest( 250U, 0U, 10, 10 );
    if ( ret != 0 ) {
	return __LINE__;
    }

    ret = runTest( 250U, 0U, 1, 1 );
    if ( ret != 0 ) {
	return __LINE__;
    }

    ret = runTest( 250U, 0U, 12, 7 );
    if ( ret != 0 ) {
	return __LINE__;
    }

    ret = runTest( 0U, 250U, 7, 12 );
    if ( ret != 0 ) {
	return __LINE__;
    }

    ret = runTest( 0U, 250U, 24, 24 );
    if ( ret != 0 ) {
	return __LINE__;
    }

    return 0;
}
#endif /* TEST */
