
#include <iostream>

#include <SDL.h>
#include <SDL_thread.h>

#include "../pipes/streamPipe.h"
#include "../elements/streamElement.h"
#include "soundDev.h"

SoundDev::SoundDev(
	StreamPipe* in,
	unsigned int freq, unsigned int channels, unsigned int samples,
	double _gain
    ) : StreamElement( in, 0 ), gain( _gain ), isFinished( false )
{
    SDL_AudioSpec desired;

    desired.freq = freq;
    desired.format = AUDIO_S16SYS;
    desired.channels = channels;
    desired.samples = samples;
    desired.callback = SoundDev::callbackTrampoline;
    desired.userdata = this;

    if ( ::SDL_OpenAudio( &desired, &this->obtained ) < 0 ) {
	std::cerr << "Couldn't open Audio:  " << ::SDL_GetError() << std::endl;
	this->gotAudio = false;
	this->obtained.samples = 0;
	this->inBuf = 0;
    } else {
	this->gotAudio = true;
	this->inBuf = new double[ this->obtained.samples ];
    }

    this->mutex = ::SDL_CreateMutex();
    this->finishedCond = ::SDL_CreateCond();
}

SoundDev::~SoundDev( void )
{
    delete[] this->inBuf;

    if ( this->gotAudio ) {
	::SDL_PauseAudio( 1 );
	::SDL_CloseAudio();
    }

    ::SDL_DestroyCond( this->finishedCond );
    ::SDL_DestroyMutex( this->mutex );
}

void
SoundDev::finish( void )
{
    ::SDL_LockMutex( this->mutex );

	if ( not this->isFinished ) {
	    ::SDL_CondWait( this->finishedCond, this->mutex );
	}

    ::SDL_UnlockMutex( this->mutex );
}

void
SoundDev::callback( Uint8* buf, unsigned int len )
{
    int rLen;

    ::SDL_LockMutex( this->mutex );

	if ( this->in not_eq 0 ) {

	    rLen = this->in->read(
		(unsigned char*)this->inBuf,
		this->obtained.samples * sizeof( this->inBuf[ 0 ] )
	    );

	    if ( rLen <= 0 ) {
		this->isFinished = true;
		this->in->readClose();
		::SDL_CondBroadcast( this->finishedCond );
		::SDL_PauseAudio( 1 );
	    } else {
		rLen /= sizeof( this->inBuf[ 0 ] );
	    }

	} else {
	    rLen = 0;
	}

	len /= sizeof( Sint16 );

	if ( rLen >= len ) {
	    rLen = len;
	}

	double* inPtr = this->inBuf;
	Sint16* ptr = (Sint16*)buf;

	for ( unsigned int ii=0; ii < rLen; ++ii ) {
	    int iVal = (int)( *inPtr++ * this->gain * 32767.0 );

	    if ( iVal < -32768 ) {
		iVal = -32768;
	    } else if ( iVal > 32767 ) {
		iVal = 32767;
	    }

	    *ptr++ = iVal;
	}

	for ( unsigned int ii=rLen; ii < len; ++ii ) {
	    *ptr++ = 0;
	}

    ::SDL_UnlockMutex( this->mutex );
}

unsigned int
SoundDev::getFreq( void ) const
{
    unsigned int ret = 0;

    ::SDL_LockMutex( this->mutex );

	if ( this->gotAudio ) {
	    ret = this->obtained.freq;
	}

    ::SDL_UnlockMutex( this->mutex );

    return ret;
}

unsigned int
SoundDev::getChannels( void ) const
{
    unsigned int ret = 0;

    ::SDL_LockMutex( this->mutex );

	if ( this->gotAudio ) {
	    return this->obtained.channels;
	}

    ::SDL_UnlockMutex( this->mutex );

    return ret;
}

int
SoundDev::run( void )
{
    ::SDL_PauseAudio( 0 );
    this->finish();
}
