
#include <iostream>
#include <stdlib.h>
#include <math.h>

#include <SDL.h>
#include <SDL_thread.h>

extern "C" double drand48( void );

#include "swarm.h"

class Bug {
public:
    Bug( bool _randomize = true );
    virtual ~Bug( void );

    void init( unsigned int _Dims, double mv, double ma, double af=1.0 );

    inline const double* getPosition( void ) const {
	return pos;
    }

    virtual void update( double elapsed, const double* target );

protected:
    bool randomize;
    unsigned int Dims;
    double* pos;
    double* vel;
    double* acc;
    double maxVelocity;
    double maxAcceleration;
    double accelFactor;
};

Bug::Bug( bool _randomize )
    : randomize( _randomize ), pos( 0 ), vel( 0 ), acc( 0 )
{
}

Bug::~Bug( void )
{
    delete[] this->acc;
    delete[] this->vel;
    delete[] this->pos;
}

void
Bug::init( unsigned int _Dims, double mv, double ma, double af )
{
    this->Dims = _Dims;
    this->maxVelocity = mv;
    this->maxAcceleration = ma;
    this->accelFactor = af;

    this->pos = new double[ Dims ];
    this->vel = new double[ Dims ];
    this->acc = new double[ Dims ];

    for ( unsigned int ii=0; ii < Dims; ++ii ) {
	this->pos[ ii ] = Swarm::randomCoord();
	this->vel[ ii ] = Swarm::randomVelocity();
	this->acc[ ii ] = Swarm::randomAcceleration();
    }
}

void
Bug::update( double elapsed, const double* target )
{
    double mag = 0.0;

    for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	this->vel[ ii ] += this->acc[ ii ] * elapsed;
	mag += this->vel[ ii ] * this->vel[ ii ];
    }

    if ( mag > this->maxVelocity * this->maxVelocity ) {
	double factor = this->maxVelocity / sqrt( mag );
	for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	    this->vel[ ii ] *= factor;
	}
    }

    for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	this->pos[ ii ] += this->vel[ ii ] * elapsed;
    }

    mag = 0.0;

    for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	this->acc[ ii ] = this->accelFactor * ( target[ ii ]-this->pos[ ii ] );

	if ( this->randomize ) {
	    this->acc[ ii ] *= Swarm::drand();
	}

	mag += this->acc[ ii ] * this->acc[ ii ];
    }

    if ( mag > this->maxAcceleration * this->maxAcceleration ) {
	double factor = this->maxAcceleration / sqrt( mag );
	for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	    this->acc[ ii ] *= factor;
	}
    }
}

Swarm::Swarm(
	unsigned int _Dims, unsigned int _GnatCount, double duration
    ) : Dims( _Dims ), GnatCount( _GnatCount ), PointDuration( duration )
{
    this->wasp = new Bug( false );
    this->wasp->init( this->Dims, 30.0, 10.0, 1.0 );

    this->gnats = new Bug[ this->GnatCount ];
    for ( unsigned int ii=0; ii < this->GnatCount; ++ii ) {
	this->gnats[ ii ].init(
	    this->Dims,
	    29.0 + randomVelocity(), 30.0 + randomAcceleration(), 1.5
	);
    }

    this->target = new double[ this->Dims ];
    this->timeUntilNextTarget = -2.0;
}

Swarm::~Swarm( void )
{
    delete[] target;
    delete[] gnats;
    delete wasp;
}

const double*
Swarm::getWaspPosition( void ) const
{
    return this->wasp->getPosition();
}

const double*
Swarm::getGnatPosition( unsigned int index ) const
{
    return this->gnats[ index ].getPosition();
}

void
Swarm::update( double elapsed )
{
    const double* waspPosition = this->wasp->getPosition();

    for ( unsigned int ii=0; ii < this->GnatCount; ++ii ) {
	this->gnats[ ii ].update( elapsed, waspPosition );
    }

    double mag = 0.0;
    for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	double dd = waspPosition[ ii ] - this->target[ ii ];
	mag += dd * dd;
    }

    this->timeUntilNextTarget -= elapsed;

    if ( mag < 100.0 or this->timeUntilNextTarget < 0.0 ) {
	for ( unsigned int ii=0; ii < this->Dims; ++ii ) {
	    this->target[ ii ] = Swarm::randomCoord() / 2.0;
	}
	this->timeUntilNextTarget = this->PointDuration;
    }

    this->wasp->update( elapsed, this->target );
}

double
Swarm::drand( void )
{
    return ::drand48();
}

double
Swarm::randomCoord( void )
{
    return Swarm::drand() * 15.0 - 7.5;
}

double
Swarm::randomVelocity( void )
{
    return Swarm::drand() * 6.0 - 3.0;
}

double
Swarm::randomAcceleration( void )
{
    return Swarm::drand() * 3.0 - 1.5;
}
